from re import sub
from string import ascii_letters, digits
from typing import Annotated

from sqlalchemy.orm import mapped_column
from sqlalchemy.types import Integer, String, TypeDecorator


class StrippedString(TypeDecorator):
    """
    A string type that strips whitespace from the ends of the string
    before sending to the database.
    """

    impl = String
    cache_ok = True

    def process_bind_param(self, string, dialect) -> str | None:
        return string.strip() if isinstance(string, str) else string


class SamplesheetString(TypeDecorator):
    """
    A string type that removes illegal characters from the string before
    sending to the database. Suitable for strings that will be used in
    the samplesheet passed as input to the `nf-tenx` pipeline.
    """

    impl = String
    cache_ok = True

    def process_bind_param(self, string: str | None, dialect) -> str | None:
        if not isinstance(string, str):
            return string

        sep_chars = r'\s_-'
        sep_chars_group = r'[\s_-]'
        samplesheet_blacklist_pattern = rf'[^{ascii_letters + digits + sep_chars}]'

        # Remove illegal characters
        string = sub(
            pattern=samplesheet_blacklist_pattern, repl='', string=string.strip()
        )

        # Replace separator characters with hyphens
        string = sub(pattern=sep_chars_group, repl='-', string=string)

        # Replace occurrences of multiple hyphens with a single hyphen
        string = sub(pattern=r'-+', repl='-', string=string)

        return string


# TODO these should just be encoded as str
# class XeniumSlideSerialNumber(TypeDecorator):
#     """ """

#     impl = Integer
#     cache_ok = True

#     def process_bind_param(self, serial_number, dialect) -> int:
#         return (
#             int(serial_number.strip())
#             if isinstance(serial_number, str)
#             else serial_number
#         )

#     def process_result_value(self, serial_number, dialect) -> str:
#         return (
#             f'{serial_number:07}' if isinstance(serial_number, int) else serial_number
#         )


# Commonly used primary key types
int_pk = Annotated[int, mapped_column(primary_key=True)]
stripped_str_pk = Annotated[str, mapped_column(StrippedString, primary_key=True)]
samplesheet_str_pk = Annotated[str, mapped_column(SamplesheetString, primary_key=True)]

# Commonly used non-unique string types
stripped_str = Annotated[str, mapped_column(StrippedString)]
samplesheet_str = Annotated[str, mapped_column(SamplesheetString)]

# Commonly used unique types
unique_int = Annotated[int, mapped_column(unique=True)]
unique_stripped_str = Annotated[str, mapped_column(StrippedString, unique=True)]
unique_samplesheet_str = Annotated[str, mapped_column(SamplesheetString, unique=True)]
# xenium_slide_serial_number = Annotated[
#     str, mapped_column(XeniumSlideSerialNumber, unique=True)
# ]



from datetime import date
from typing import ClassVar, Literal

from sqlalchemy import ForeignKey, ForeignKeyConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from ..bases import Base, Data
from ..custom_types import samplesheet_str, stripped_str
from .entities import Lab, Person
from .processes import Assay

__all__ = ['DataSet', 'Sample']


class DataSet(Data, kw_only=True):
    __tablename__ = 'data_set'

    # DataSet attributes
    # TODO: ilab request ID validation
    ilab_request_id: Mapped[stripped_str] = mapped_column(index=True, repr=False)
    date_initialized: Mapped[date] = mapped_column(repr=False)

    # Parent foreign keys
    assay_name: Mapped[str] = mapped_column(
        ForeignKey('assay.name'), init=False, repr=False
    )
    lab_name: Mapped[str] = mapped_column(init=False, repr=False)
    lab_pi_email: Mapped[str] = mapped_column(init=False, repr=False)
    lab_institution_name: Mapped[str] = mapped_column(init=False, repr=False)
    platform_name: Mapped[str] = mapped_column(
        ForeignKey('platform.name'), init=False, repr=False
    )
    submitter_email: Mapped[str] = mapped_column(
        ForeignKey('person.email'), init=False, repr=False
    )

    # Parent models
    assay: Mapped[Assay] = relationship()
    lab: Mapped[Lab] = relationship()
    submitter: Mapped[Person] = relationship()

    # Model metadata
    id_date_col: ClassVar[Literal['date_initialized']] = 'date_initialized'

    __mapper_args__ = {
        'polymorphic_on': 'platform_name',
    }

    __table_args__ = (
        ForeignKeyConstraint(
            columns=['lab_name', 'lab_pi_email', 'lab_institution_name'],
            refcolumns=['lab.name', 'lab.pi_email', 'lab.institution_name'],
        ),
    )


class Sample(Data, kw_only=True):
    __tablename__ = 'sample'

    # Sample attributes
    name: Mapped[samplesheet_str] = mapped_column(index=True)
    date_received: Mapped[date] = mapped_column(repr=False)

    # Parent foreign keys
    data_set_id: Mapped[str] = mapped_column(
        ForeignKey('data_set.id'), default=None, init=False, repr=False
    )
    platform_name: Mapped[str] = mapped_column(
        ForeignKey('platform.name'), init=False, repr=False
    )

    # Model metadata
    id_date_col: ClassVar[Literal['date_received']] = 'date_received'

    __mapper_args__ = {'polymorphic_on': 'platform_name'}



from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..bases import Process

__all__ = ['Platform', 'Assay']


class Platform(Process, kw_only=True):
    __tablename__ = 'platform'

    assays: Mapped[list['Assay']] = relationship(
        back_populates='platform', default_factory=list
    )


class Assay(Process, kw_only=True):
    __tablename__ = 'assay'

    # Parent foreign keys
    platform_name: Mapped[str] = mapped_column(
        ForeignKey('platform.name'), default=None, init=False, repr=False
    )

    platform: Mapped[Platform] = relationship(back_populates='assays')



from datetime import date
from typing import ClassVar, Literal

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from ...bases import Data
from ...custom_types import samplesheet_str, stripped_str
from ..data import DataSet, Sample

__all__ = ['XeniumRun', 'XeniumDataSet', 'XeniumRegion']


# TODO: This set of models is incomplete
class XeniumRun(Data):
    __tablename__ = 'xenium_run'

    # XeniumRun attributes
    date_begun: Mapped[date] = mapped_column(repr=False)

    # Model metadata
    id_date_col: ClassVar[Literal['date_begun']] = 'date_begun'
    id_prefix: ClassVar[Literal['XR']] = 'XR'
    id_length: ClassVar[Literal[7]] = 7

    # Child models
    data_sets: Mapped[list['XeniumDataSet']] = relationship(
        back_populates='xenium_run', default_factory=list, repr=False, compare=False
    )


class XeniumDataSet(DataSet, kw_only=True):
    # XeniumDataSet attributes
    slide_serial_number: Mapped[stripped_str | None]
    slide_name: Mapped[samplesheet_str | None]

    # Parent foreign keys
    xenium_run_id: Mapped[int | None] = mapped_column(
        ForeignKey('xenium_run.id'), default=None, repr=False, init=False
    )

    # Parent models
    xenium_run: Mapped[XeniumRun] = relationship(
        back_populates='data_sets', default=None
    )

    # Child models
    samples: Mapped[list['XeniumRegion']] = relationship(
        back_populates='data_set', default_factory=list, repr=False, compare=False
    )

    # Model metadata
    id_prefix: ClassVar[Literal['XD']] = 'XD'
    id_length: ClassVar[Literal[8]] = 8

    __mapper_args__ = {'polymorphic_identity': 'Xenium'}

    # TODO: implement this to check against 10x's database?
    @validates('slide_serial_number')
    def check_slide_serial_number(self, key: str, serial_number: str) -> str:
        serial_number = serial_number.strip()

        try:
            int(serial_number)
        except ValueError:
            raise ValueError(f'{key} must be a string of numbers.')

        correct_serial_number_length = 7
        if len(serial_number) != correct_serial_number_length:
            raise ValueError(
                f'{key} must be {correct_serial_number_length} characters long.'
            )

        return serial_number


class XeniumRegion(Sample):
    # Parent models
    data_set: Mapped[XeniumDataSet] = relationship(
        back_populates='samples', default=None
    )

    # Model metadata
    id_prefix: ClassVar[Literal['XE']] = 'XE'
    id_length: ClassVar[Literal[8]] = 8

    __mapper_args__ = {'polymorphic_identity': 'Xenium'}
